from yta_image.parser import ImageParser
from yta_general_utils.temp import Temp
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.dataclasses import FileReturn
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from PIL import Image
from typing import Union
from subprocess import run

import numpy as np


class ImageBackgroundRemover:
    """
    Class to remove backgrounds from images.
    """

    @staticmethod
    def remove_background(
        image: Union[str, Image.Image, np.ndarray],
        output_filename: Union[str, None] = None
    ) -> FileReturn:
        """
        Remove the background of the provided 'image'. This
        method returns the image as a pillow image in the
        first element, and the created image filename as the
        second element.
        """
        image_filename = image

        # If provided image is not a file, we store it
        # to be able to handle
        if not PythonValidator.is_string(image):
            image_filename = Temp.create_filename('background_to_remove.png')
            ImageParser.to_pillow(image).save(image_filename)

        output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)

        command_parameters = ['backgroundremover', '-i', image_filename, '-o', output_filename]

        run(command_parameters)

        return FileReturn(
            ImageParser.to_pillow(output_filename),
            output_filename
        )
    
    """
    # Problem with Circular import
    from backgroundremover.bg import remove as remove_background
    r = lambda image_filename: image_filename.buffer.read() if hasattr(image_filename, "buffer") else image_filename.read()
    w = lambda o, data: o.buffer.write(data) if hasattr(o, "buffer") else o.write(data)

    # These below are default values
    x = remove_background(
        r(image_filename),
        model_name = 'u2net',
        alpha_matting = False,
        alpha_matting_foreground_threshold = 240,
        alpha_matting_background_threshold = 10,
        alpha_matting_erode_structure_size = 10,
        alpha_matting_base_size = 1000
    )
    w(output_filename, x)
    """

    # TODO: This below seems to work (as shown in this 
    # commit https://github.com/nadermx/backgroundremover/commit/c590858de4c7e75805af9b8ecdd22baf03a1368f)
    """
    from backgroundremover.bg import remove
    def remove_bg(src_img_path, out_img_path):
        model_choices = ["u2net", "u2net_human_seg", "u2netp"]
        f = open(src_img_path, "rb")
        data = f.read()
        img = remove(data, model_name=model_choices[0],
                    alpha_matting=True,
                    alpha_matting_foreground_threshold=240,
                    alpha_matting_background_threshold=10,
                    alpha_matting_erode_structure_size=10,
                    alpha_matting_base_size=1000)
        f.close()
        f = open(out_img_path, "wb")
        f.write(img)
        f.close()
    """


from yta_ai_utils.blip import Blip
from yta_ai_utils.llava import Llava
from yta_general_utils.programming.validator import PythonValidator
from PIL import Image
from typing import Union
from abc import ABC, abstractmethod

import numpy as np


class ImageDescriptor(ABC):
    """
    Class to describe images.
    """

    @abstractmethod
    def describe(
        self,
        image: Union[str, Image.Image, np.ndarray]
    ) -> str:
        """
        Describe the provided 'image' using an engine
        capable of it.
        """
        pass

class DefaultImageDescriptor(ImageDescriptor):
    """
    Default class to describe an image. It will choose the
    engine we think is a good choice in general.

    The process could take from seconds to a couple of minutes
    according to the system specifications.
    """
    def describe(
        self,
        image: Union[str, Image.Image, np.ndarray]
    ) -> str:
        return BlipImageDescriptor().describe(image)

class BlipImageDescriptor(ImageDescriptor):
    """
    Class to describe an image using the Blip engine, which
    is from Salesforce and will use pretrained models that are
    stored locally in 'C:/Users/USER/.cache/huggingface/hub',
    loaded in memory and used to describe it.

    The process could take from seconds to a couple of minutes
    according to the system specifications.
    """

    def describe(
        self,
        image: Union[str, Image.Image, np.ndarray]
    ) -> str:
        return Blip.describe(image)
    
class LlavaImageDescriptor(ImageDescriptor):
    """
    Class to describe an image using the Llava engine
    through the 'ollama' python package.
    """

    def describe(
        self,
        image_filename: str
    ):
        """
        THIS METHOD IS NOT WORKING YET.

        TODO: This is not working because of my pc limitations.
        It cannot load the resources due to memory capacity.
        """
        if not PythonValidator.is_string(image_filename):
            raise Exception('The provided "image_filename" is not a valid string.')

        return Llava.describe(image_filename)


from yta_image.parser import ImageParser
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileExtension
from yta_general_utils.dataclasses import FileReturn
from typing import Union
from PIL import Image


def pixelate_image(
    image,
    i_size,
    output_filename: Union[str, None] = None
) -> FileReturn:
    """
    Pixelates the provided 'image' and saves it as the 'output_filename' if
    provided.
    The 'i_size' is the pixelating square. The smaller it is, the less pixelated 
    its.

    'i_size' must be a tuple such as (8, 8) or (16, 16).
    """
    img = ImageParser.to_pillow(image)

    # Convert to small image
    small_img = img.resize(i_size, Image.BILINEAR)

    # Resize to output size
    res = small_img.resize(img.size, Image.NEAREST)

    output_filename = Output.get_filename(output_filename, FileExtension.PNG)
        
    return FileReturn(
        res,
        output_filename
    )


from yta_image.converter import ImageConverter
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.file.filename import filename_is_type
from yta_general_utils.file.enums import FileType
from yta_general_utils.programming.validator import PythonValidator
from PIL import Image
from typing import Union

import numpy as np


class ImageParser:
    """
    Class to simplify the way we handle the image parameters
    so we can parse them as Pillow images, as numpy arrays,
    etc.
    """

    @staticmethod
    def to_pillow(
        image: Union[str, Image.Image, np.ndarray],
        mode: str = 'RGB'
    ) -> Image.Image:
        """
        Returns an instance of a Pillow Image.Image of the given
        'image' if it is a valid image and no error found.

        Result is a Pillow image which is in RGB (or RGBA) format.
        """
        # TODO: By now we are only accepting string filenames,
        # Pillow Image.Image instances and numpy arrays.
        if image is None:
            raise Exception('No "image" parameter provided.')
        
        if (
            not PythonValidator.is_string(image) and
            not PythonValidator.is_numpy_array(image) and
            not PythonValidator.is_instance(image, Image.Image)
        ):
            raise Exception('The "image" parameter provided is not a string nor a Image.Image nor a np.ndarray.')

        mode = 'RGB' if not mode else mode

        if mode not in ['RGB', 'RGBA']:
            raise Exception('The provided "mode" parameters is not a valid mode: RGB or RGBA.')

        # We can have problems with np.ndarray
        if PythonValidator.is_numpy_array(image):
            image = ImageConverter.numpy_image_to_pil(image)
        elif PythonValidator.is_string(image):
            if not filename_is_type(image, FileType.IMAGE):
                raise Exception('The "image" parameter provided is not a valid image filename.')
            
            if not FileValidator.file_is_image_file(image):
                raise Exception('The "image" parameter provided is not a valid image.')
            
            image = Image.open(image)

        return image.convert(mode)
    
    @staticmethod
    def to_numpy(
        image: Union[str, Image.Image, np.ndarray],
        mode: str = 'RGB'
    ) -> np.ndarray:
        """
        Returns a numpy array representing the given 'image'
        if it is a valid image and no error found.

        The 'mode' parameter will be used to open the image
        with Pillow library and then turning into numpy when
        necessary. Must be 'RGB' or 'RGBA'.

        Result is in RGB or RGBA format.
        """
        # TODO: By now we are only accepting string filenames,
        # Pillow Image.Image instances and numpy arrays.
        if image is None:
            raise Exception('No "image" parameter provided.')
        
        if not PythonValidator.is_instance(image, [str, Image.Image, np.ndarray]):
            raise Exception('The "image" parameter provided is not a string nor a Image.Image nor a np.ndarray.')
        
        mode = 'RGB' if not mode else mode

        if mode not in ['RGB', 'RGBA']:
            raise Exception('The provided "mode" parameters is not a valid mode: RGB or RGBA.')

        # We can have problems with np.ndarray
        if PythonValidator.is_instance(image, Image.Image):
            image = ImageConverter.pil_image_to_numpy(image.convert(mode))
        elif PythonValidator.is_string(image):
            if not filename_is_type(image, FileType.IMAGE):
                raise Exception('The "image" parameter provided is not a valid image filename.')
            
            if not FileValidator.file_is_image_file(image):
                raise Exception('The "image" parameter provided is not a valid image.')
            
            image = ImageConverter.pil_image_to_numpy(Image.open(image).convert(mode))
        elif PythonValidator.is_numpy_array(image):
            if image.shape[2] == 3 and mode == 'RGBA':
                # numpy but RGB to RGBA
                image = np.dstack((image, np.ones((image.shape[0], image.shape[1]), dtype = np.uint8) * 255))
                # Moviepy uses the alpha channel as 0 or 1, not as 255
                # but this is only an image parser not a moviepy parser
                #image = np.dstack((image, np.ones((image.shape[0], image.shape[1]), dtype = np.uint8)))
            elif image.shape[2] == 4 and mode == 'RGB':
                # numpy but RGBA to RGB
                image = image[:, :, :3]

        return image
    
    @staticmethod
    def to_opencv(
        image: Union[str, Image.Image, np.ndarray],
        mode: str = 'RGB'
    ) -> np.ndarray:
        """
        The 'image' is read as a RGB numpy array and then
        transformed into an BGR numpy array because Opencv
        uses BGR format.

        Result is in BGR format.
        """
        return ImageParser.to_numpy(image, mode)[:, :, ::-1]